<!doctype html><html lang="en"><head><meta charset="utf-8"/><link rel="icon" href="{{ base_path }}/static-files/favicon.ico"/><meta name="viewport" content="minimum-scale=1,initial-scale=1,width=device-width"/><meta name="theme-color" content="#000000"/><meta name="description" content="Open-source, self-hosted AI experiment tracking tool"/><link rel="manifest" href="{{ base_path }}/static-files/manifest.json"/><link rel="stylesheet" href="https://fonts.googleapis.com/icon?family=Material+Icons"/><link rel="stylesheet" href="{{ base_path }}/static-files/assets/icomoon/icomoonIcons.css"/><link rel="stylesheet" href="{{ base_path }}/static-files/assets/inter/inter.css"/><title>Aim</title><script>!function(){var e=window.analytics=window.analytics||[];if(!e.initialize)if(e.invoked)window.console&&console.error&&console.error("Segment snippet included twice.");else{e.invoked=!0,e.methods=["trackSubmit","trackClick","trackLink","trackForm","pageview","identify","reset","group","track","ready","alias","debug","page","once","off","on","addSourceMiddleware","addIntegrationMiddleware","setAnonymousId","addDestinationMiddleware"],e.factory=function(t){return function(){var n=Array.prototype.slice.call(arguments);return n.unshift(t),e.push(n),e}};for(var t=0;t<e.methods.length;t++){var n=e.methods[t];e[n]=e.factory(n)}e.load=function(t,n){var a=document.createElement("script");a.type="text/javascript",a.async=!0,a.src="https://cdn.segment.com/analytics.js/v1/"+t+"/analytics.min.js";var r=document.getElementsByTagName("script")[0];r.parentNode.insertBefore(a,r),e._loadOptions=n},e.SNIPPET_VERSION="4.15.3"}}()</script><script>window.API_BASE_PATH="{{ base_path }}",window.externalPublicPath="{{ base_path }}/static-files/"</script><link href="{{ base_path }}/static-files/static/css/main.62b108cb.chunk.css" rel="stylesheet"></head><body><noscript>You need to enable JavaScript to run this app.</noscript><div id="root"></div><script src="{{ base_path }}/static-files/static/js/runtime~main.c8842a07.js"></script><script src="{{ base_path }}/static-files/static/js/main.5aad559d.chunk.js"></script></body></html>